"""
The wntr.network.base module includes base classes for network elements and 
the network model.

.. rubric:: Contents

.. autosummary::

    AbstractModel
    Subject
    Observer
    Node
    Link
    Registry
    NodeType
    LinkType
    LinkStatus

"""
import logging
import six
from six import string_types
import types
from wntr.utils.ordered_set import OrderedSet

import enum
import sys
if sys.version_info[0] == 2:
    from collections import MutableSequence, MutableMapping
else:
    from collections.abc import MutableSequence, MutableMapping
from collections import OrderedDict
from wntr.utils.ordered_set import OrderedSet

import abc

logger = logging.getLogger(__name__)


class AbstractModel(object):
    """
    Base class for water network models.
    """
    pass


class Subject(object):
    """
    Base class for the subject in an observer design pattern.
    """
    def __init__(self):
        self._observers = OrderedSet()

    def subscribe(self, observer):
        self._observers.add(observer)

    def unsubscribe(self, observer):
        self._observers.remove(observer)

    def notify(self):
        for o in self._observers:
            o.update(self)


class Observer(six.with_metaclass(abc.ABCMeta, object)):
    """
    Base class for the observer in an observer design pattern.
    """
    @abc.abstractmethod
    def update(self, subject):
        pass


class Node(six.with_metaclass(abc.ABCMeta, object)):
    """Base class for nodes.
    
    For details about the different subclasses, see one of the following:
    :class:`~wntr.network.elements.Junction`, 
    :class:`~wntr.network.elements.Tank`, and
    :class:`~wntr.network.elements.Reservoir`

    Parameters
    -----------
    wn : :class:`~wntr.network.model.WaterNetworkModel`
        WaterNetworkModel object
    name : string
        Name of the node (must be unique among nodes of all types)


    """
    def __init__(self, wn, name):
        self._name = name
        self._head = None
        self._demand = None  
        self._leak_demand = None
        self._initial_quality = None
        self._tag = None
        self._leak = False
        self._leak_status = False
        self._leak_area = 0.0
        self._leak_discharge_coeff = 0.0
        self._options = wn._options
        self._node_reg = wn._node_reg
        self._link_reg = wn._link_reg
        self._controls = wn._controls
        self._pattern_reg = wn._pattern_reg
        self._curve_reg = wn._curve_reg
        self._coordinates = [0,0]
        self._source = None
        self._is_isolated = False

    def _compare(self, other):
        if not type(self) == type(other):
            return False
        if self.name == other.name and \
           self.initial_quality == other.initial_quality and \
           self.tag == other.tag:
               return True
        return False

    def __str__(self):
        return self._name

    def __repr__(self):
        return "<Node '{}'>".format(self._name)

    @property
    def head(self):
        """float: The current head at the node"""
        return self._head
    @head.setter
    def head(self, value):
        self._head = value

    @property
    def demand(self):
        """float: The current demand at the node"""
        return self._demand
    @demand.setter
    def demand(self, value):
        self._demand = value

    @property
    def leak_demand(self):
        """float: The current demand at the node"""
        return self._leak_demand
    @leak_demand.setter
    def leak_demand(self, value):
        self._leak_demand = value

    @property
    def leak_status(self):
        """bool: The current leak status at the node"""
        return self._leak_status
    @leak_status.setter
    def leak_status(self, value):
        self._leak_status = value

    @property
    def leak_area(self):
        """float: The leak area at the node"""
        return self._leak_area
    @leak_area.setter
    def leak_area(self, value):
        self._leak_area = value

    @property
    def leak_discharge_coeff(self):
        """float: The leak discharge coefficient"""
        return self._leak_discharge_coeff
    @leak_discharge_coeff.setter
    def leak_discharge_coeff(self, value):
        self._leak_discharge_coeff = value

    @property
    def node_type(self):
        """str: The node type"""
        return 'Node'
    
    @property
    def name(self):
        """str: The name of the node"""
        return self._name
    
    @property
    def tag(self):
        """str: A tag or label for the node"""
        return self._tag
    @tag.setter
    def tag(self, tag):
        self._tag = tag

    @property
    def initial_quality(self):
        """float: The initial quality (concentration) at the node"""
        if not self._initial_quality:
            return 0.0
        return self._initial_quality
    @initial_quality.setter
    def initial_quality(self, value):
        if value and not isinstance(value, (list, float, int)):
            raise ValueError('Initial quality must be a float or a list')
        self._initial_quality = value

    @property
    def coordinates(self):
        """tuple: The node coordinates, (x,y)"""
        return self._coordinates
    @coordinates.setter
    def coordinates(self, coordinates):
        if isinstance(coordinates, (list, tuple)) and len(coordinates) == 2:
            self._coordinates = tuple(coordinates)
        else:
            raise ValueError('coordinates must be a 2-tuple or len-2 list')

    def todict(self):
        """Dictionary representation of the node"""
        d = {}
        for k in dir(self):
            if not k.startswith('_'):
                val = getattr(self, k)
                if not isinstance(val, types.MethodType):
                    d[k] = val
        return d


class Link(six.with_metaclass(abc.ABCMeta, object)):
    """Base class for links.

    For details about the different subclasses, see one of the following:
    :class:`~wntr.network.elements.Pipe`, 
    :class:`~wntr.network.elements.Pump`, and
    :class:`~wntr.network.elements.Valve`

    Parameters
    ----------
    wn : :class:`~wntr.network.model.WaterNetworkModel`
        WaterNetworkModel object
    link_name : string
        Name of the link
    start_node_name : string
        Name of the start node
    end_node_name : string
        Name of the end node
    
    """
    def __init__(self, wn, link_name, start_node_name, end_node_name):
        # Set the registries
        self._options = wn._options
        self._node_reg = wn._node_reg
        self._link_reg = wn._link_reg
        self._controls = wn._controls
        self._pattern_reg = wn._pattern_reg
        self._curve_reg = wn._curve_reg
        # Set the link name
        self._link_name = link_name
        # Set and register the starting node
        self._start_node = self._node_reg[start_node_name]
        self._node_reg.add_usage(start_node_name, (link_name, self.link_type))
        # Set and register the ending node
        self._end_node = self._node_reg[end_node_name]
        self._node_reg.add_usage(end_node_name, (link_name, self.link_type))
        # Set up other metadata fields
        self._initial_status = LinkStatus.opened
        self._initial_setting = None
        self._vertices = []
        self._tag = None
        # Model state variables
        self._user_status = LinkStatus.opened
        self._internal_status = LinkStatus.active
        self._prev_setting = None
        self._setting = None
        self._flow = None
        self._is_isolated = False

    def _compare(self, other):
        """
        Parameters
        ----------
        other: Link

        Returns
        -------
        bool
        """
        if not type(self) == type(other):
            return False
        if self.name != other.name:
            return False
        if self.tag != other.tag:
            return False
        if self.initial_status != other.initial_status:
            return False
        if self.initial_setting != other.initial_setting:
            return False
        if self.start_node_name != other.start_node_name:
            return False
        if self.end_node_name != other.end_node_name:
            return False
        return True
    
    def __str__(self):
        return self._link_name

    def __repr__(self):
        return "<Link '{}'>".format(self._link_name)

    @property
    def link_type(self):
        """str: The link type"""
        return 'Link'

    @property
    def initial_status(self):
        """:class:`~wntr.network.base.LinkStatus`: The initial status (`Opened`, `Closed`, `Active`) of the Link"""
        return self._initial_status
    @initial_status.setter
    def initial_status(self, status):
        if not isinstance(status, LinkStatus):
            status = LinkStatus[status]
        self._initial_status = status
        
    @property
    def initial_setting(self):
        """float: The initial setting for the link (if `Active`)"""
        return self._initial_setting
    @initial_setting.setter
    def initial_setting(self, setting):
        # TODO: typechecking
        self._initial_setting = setting

    @property
    def start_node(self):
        """:class:`~wntr.network.base.Node`: The start node object."""
        return self._start_node
    @start_node.setter
    def start_node(self, node):
        self._node_reg.remove_usage(self.start_node_name, (self._link_name, self.link_type))
        self._node_reg.add_usage(node.name, (self._link_name, self.link_type))
        self._start_node = self._node_reg[node.name]

    @property
    def end_node(self):
        """:class:`~wntr.network.base.Node`: The end node object."""
        return self._end_node
    @end_node.setter
    def end_node(self, node):
        self._node_reg.remove_usage(self.end_node_name, (self._link_name, self.link_type))
        self._node_reg.add_usage(node.name, (self._link_name, self.link_type))
        self._end_node = self._node_reg[node.name]

    @property
    def start_node_name(self):
        """str: The name of the start node (read only)"""
        return self._start_node.name
    
    @property
    def end_node_name(self):
        """str: The name of the end node (read only)"""
        return self._end_node.name

    @property
    def name(self):
        """str: The link name (read-only)"""
        return self._link_name

    @property
    def flow(self):
        """float: Current flow through the link (read only)"""
        return self._flow
    
    @property
    @abc.abstractmethod
    def status(self):
        """:class:`~wntr.network.base.LinkStatus`: Current status of the link"""
        pass
    @status.setter
    @abc.abstractmethod
    def status(self, status):
        self._user_status = status
    
    @property
    def setting(self):
        """float: The current setting of the link"""
        return self._setting
    @setting.setter
    def setting(self, setting):
        self._setting = setting
    
    @property
    def tag(self):
        """str: A tag or label for this link"""
        return self._tag
    @tag.setter
    def tag(self, tag):
        self._tag = tag
        
    @property
    def vertices(self):
        """A list of curve points, in the direction of start node to end node.
        
        The vertices should be listed as a list of (x,y) tuples when setting.
        """
        return self._vertices
    @vertices.setter
    def vertices(self, points):
        if not isinstance(points, list):
            raise ValueError('vertices must be a list of 2-tuples')
        for pt in points:
            if not isinstance(pt, tuple) or len(pt) != 2:
                raise ValueError('vertices must be a list of 2-tuples')
        self._vertices = points
    
    def todict(self):
        """Dictionary representation of the link"""
        d = {}
        for k in dir(self):
            if not k.startswith('_'):
                val = getattr(self, k)
                if not isinstance(val, types.MethodType):
                    d[k] = val
        return d


class Registry(MutableMapping):
    """Base class for registries.
    
    Parameters
    ----------
    wn : :class:`~wntr.network.model.WaterNetworkModel`
        WaterNetworkModel object
    """
    
    def __init__(self, wn):
        if not isinstance(wn, AbstractModel):
            raise ValueError('Registry must be initialized with a model')
#        self._m = model
        self._data = OrderedDict()
        self._usage = OrderedDict()

    def _finalize_(self, wn):
        self._options = wn._options
        self._pattern_reg = wn._pattern_reg
        self._curve_reg = wn._curve_reg
        self._node_reg = wn._node_reg
        self._link_reg = wn._link_reg
        self._controls = wn._controls
        self._sources = wn._sources
    
    def __getitem__(self, key):
        if not key:
            return None
        try:
            return self._data[key]
        except KeyError:
            try:
                return self._data[key.name]
            except:
                return self._data[str(key)]
            

    def __setitem__(self, key, value):
        if not isinstance(key, string_types):
            raise ValueError('Registry keys must be strings')
        self._data[key] = value
    
    def __delitem__(self, key):
        try:
            if self._usage and key in self._usage and len(self._usage[key]) > 0:
                raise RuntimeError('cannot remove %s %s, still used by %s', 
                                   self.__class__.__name__,
                                   key,
                                   self._usage[key])
            elif key in self._usage:
                self._usage.pop(key)
            return self._data.pop(key)
        except KeyError:
            # Do not raise an exception if there is no key of that name
            return
    
    def __iter__(self):
        return self._data.__iter__()
    
    def __len__(self):
        return len(self._data)

    def __call__(self):
        for key, value in self._data.items():
            yield key, value

    def usage(self):
        """Generator to get the usage for all objects in the registry
        
        Yields
        ------
        key : str
            The name of the object in the registry
        value : tuple of (str, str)
            Tuple of (name, typestr) of the external items using the object
        """
        for k, v in self._usage.items():
            yield k, v
        
    def get_usage(self, key):
        """Get a set of items using an object by key.
        
        Returns
        -------
        set of 2-tuples
            Set of (name, typestr) of the external object using the item
            
        """
        try:
            return self._usage[key]
        except KeyError:
            try:
                return self._usage[str(key)]
            except KeyError:
                return None
        return None

    def orphaned(self):
        """Get a list of orphaned usages.
        
        If removed without appropriate checks, it is possible that a some other 
        item will point to an object that has been deleted. (This is why the user
        should always use the "remove_*" methods). This method returns a list of 
        names for objects that are referenced, but no longer exist.
        
        Returns
        -------
        set
            The names of any orphaned items
        
        """
        defined = set(self._data.keys())
        assigned = set(self._usage.keys())
        return assigned.difference(defined)
    
    def unused(self):
        """Get a list of items which are unused by other objects in the model.
        
        In most cases, this method will give little information. For nodes, however,
        this method could be important to identify a node which has become completely 
        disconnected from the network. For patterns or curves, it may be used to find
        extra patterns or curves that are no longer necessary (or which the user 
        forgot ot assign). It is not terribly useful for links.
        
        Returns
        -------
        set
            The names of any unused objects in the registry
            
        """
        defined = set(self._data.keys())
        assigned = set(self._usage.keys())
        return defined.difference(assigned)

    def clear_usage(self, key):
        """if key in usage, clear usage[key]"""
        if not key:
            return
        self._usage[key].clear()
    
    def add_usage(self, key, *args):
        """add args to usage[key]"""
        if not key:
            return
        if not (key in self._usage):
            self._usage[key] = OrderedSet()
        for arg in args:
            self._usage[key].add(arg)

    def remove_usage(self, key, *args):
        """remove args from usage[key]"""
        if not key:
            return
        for arg in args:
            self._usage[key].discard(arg)
        if len(self._usage[key]) < 1:
            self._usage.pop(key)

    def todict(self):
        """Dictionary representation of the registry"""
        d = dict()
        for k, v in self._data.items():
            d[k] = v.todict()
        return d
    
    def tolist(self):
        """List representation of the registry"""
        l = list()
        for k, v in self._data.items():
            l.append(v.todict())
        return l


class NodeType(enum.IntEnum):
    """
    Enum class for node types.

    .. rubric:: Enum Members

    ==================  ==================================================================
    :attr:`~Junction`   Node is a :class:`~wntr.network.model.Junction`
    :attr:`~Reservoir`  Node is a :class:`~wntr.network.model.Reservoir`
    :attr:`~Tank`       Node is a :class:`~wntr.network.model.Tank`
    ==================  ==================================================================

    """
    Junction = 0
    Reservoir = 1
    Tank = 2

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or \
               self.__class__.__name__ == other.__class__.__name__)


class LinkType(enum.IntEnum):
    """
    Enum class for link types.

    .. rubric:: Enum Members

    ===============  ==================================================================
    :attr:`~CV`      Pipe with check valve
    :attr:`~Pipe`    Regular pipe
    :attr:`~Pump`    Pump
    :attr:`~Valve`   Any valve type (see following)
    :attr:`~PRV`     Pressure reducing valve
    :attr:`~PSV`     Pressure sustaining valve
    :attr:`~PBV`     Pressure breaker valve
    :attr:`~FCV`     Flow control valve
    :attr:`~TCV`     Throttle control valve
    :attr:`~GPV`     General purpose valve
    ===============  ==================================================================

    """
    CV = 0
    Pipe = 1
    Pump = 2
    PRV = 3
    PSV = 4
    PBV = 5
    FCV = 6
    TCV = 7
    GPV = 8
    Valve = 9

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or \
               self.__class__.__name__ == other.__class__.__name__)


class LinkStatus(enum.IntEnum):
    """
    Enum class for link statuses.

    .. warning::
        This is NOT the class for determining output status from an EPANET binary file.
        The class for output status is wntr.epanet.util.LinkTankStatus.

    .. rubric:: Enum Members

    =================  ==================================================================
    :attr:`~Closed`    Pipe/valve/pump is closed.
    :attr:`~Opened`    Pipe/valve/pump is open.
    :attr:`~Open`      Alias to "Opened"
    :attr:`~Active`    Valve is partially open.
    =================  ==================================================================

    """
    Closed = 0
    Open = 1
    Opened = 1
    Active = 2
    CV = 3

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or
                                            self.__class__.__name__ == other.__class__.__name__)


class _DemandStatus(enum.IntEnum):
    """
    Enum class for demand statuses.

    .. warning::
        This is NOT the class for determining output status from an EPANET binary file.
        The class for output status is wntr.epanet.util.LinkTankStatus.

    .. rubric:: Enum Members

    ================  ==================================================================
    :attr:`~Full`     Full demand is provided.
    :attr:`~Partial`  Partial demand is provided.
    :attr:`~Zero`     No demand is provided.
    ================  ==================================================================

    """
    Full = 0
    Partial = 1
    Zero = 1

    def __init__(self, val):
        if self.name != self.name.upper():
            self._member_map_[self.name.upper()] = self
        if self.name != self.name.lower():
            self._member_map_[self.name.lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or \
                                            self.__class__.__name__ == other.__class__.__name__)




"""
The wntr.graphics.curve module includes methods plot fragility curves and 
pump curves.
"""
import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    plt = None
import logging

logger = logging.getLogger(__name__)

def plot_fragility_curve(FC, fill=True, key='Default',
                         title='Fragility curve', 
                         xmin=0, xmax=1, npoints=100, 
                         xlabel='x', 
                         ylabel='Probability of exceeding a damage state',
                         figsize=[8,4]):
    """
    Plot fragility curve.
    
    Parameters
    -----------
    FC : wntr.scenario.FragilityCurve object
        Fragility curve
    
    fill : bool (optional)
        If true, fill area under the curve (default = True)
    
    key : string (optional)
        Fragility curve state distribution key (default = 'Default')
    
    title : string (optional)
        Plot title
    
    xmin : float (optional)
        X axis minimum (default = 0)
    
    xmax : float (optional)
        X axis maximum (default = 1)
    
    npoints : int (optional)
        Number of points (default = 100)
    
    xlabel : string (optional)
        X axis label (default = 'x')
    
    ylabel : string (optional)
        Y axis label (default = 'Probability of exceeding a damage state')
    
    figsize : list (optional)
        Figure size (default = [8,4])
"""
    if plt is None:
        raise ImportError('matplotlib is required')
    
    plt.figure(figsize=tuple(figsize))
    plt.title(title)
    x = np.linspace(xmin,xmax,npoints)
    for name, state in FC.states():
        try:
            dist=state.distribution[key]
            if fill:
                plt.fill_between(x,dist.cdf(x), label=name)
            else:
                plt.plot(x,dist.cdf(x), label=name)
        except:
            pass        
    plt.xlim((xmin,xmax))
    plt.ylim((0,1))
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.legend()

def plot_pump_curve(pump, title='Pump curve', 
                    xmin=0, xmax=None, ymin=0, ymax=None, 
                    xlabel='Head (m)', 
                    ylabel='Flow (m3/s)',
                    figsize=[8,4]):
    """
    Plot pump curve.
    
    Parameters
    -----------
    pump : wntr.scenario.Pump object
        Pump
        
    title : string (optional)
        Plot title
    
    xmin : float (optional)
        X axis minimum (default = 0)
    
    xmax : float (optional)
        X axis maximum (default = None)
    
    ymin : float (optional)
        Y axis minimum (default = 0)
    
    ymax : float (optional)
        Y axis maximum (default = None)
    
    xlabel : string (optional)
        X axis label (default = 'Head (m)')
    
    ylabel : string (optional)
        Y axis label (default = 'Flow (m3/s)')
    
    figsize : list (optional)
        Figure size (default = [8,4])
    """
    try:
        curve = pump.get_pump_curve()
    except:
        print("Pump "+pump.name+" has no curve")
        return

    if plt is None:
        raise ImportError('matplotlib is required')
    
    plt.figure(figsize=tuple(figsize))
    plt.title(title)
    x = []
    y = []
    for pt in curve.points:
        x.append(pt[0])
        y.append(pt[1])
    plt.scatter(x,y)
    plt.plot(x,y, label=curve.name)    
    if xmax is None:
        xmax = max(x)
    if ymax is None:
        ymax = max(y)
    plt.xlim((xmin,xmax))
    plt.ylim((ymin,ymax))
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.legend()
    



"""
The wntr.network.graph module includes methods to represent a water network 
model as a NetworkX MultiDiGraph, and compute topographic metrics on the graph.

"""
import networkx as nx
import numpy as np
import pandas as pd
import logging

logger = logging.getLogger(__name__)

class WntrMultiDiGraph(nx.MultiDiGraph):
    """
    Extension of networkx MultiDiGraph
    """

    def weight_graph(self, node_attribute={}, link_attribute={}):
        """
        Return a weighted graph based on node and link attributes.
        The weighted graph changes the direction of the original link if the weight is negative.

        Parameters
        ----------
        G : graph
            Networkx graph
        node_attribute :  dict or pandas Series
            Node attributes
        link_attribues : dict or pandas Series
            Link attributes

        Returns
        -------
        Networkx weighted graph
        
        """

        for node_name in self.nodes():
            try:
                value = node_attribute[node_name]

                nx.set_node_attributes(self, name='weight', values={node_name: value})
            except:
                pass

        for (node1, node2, link_name) in list(self.edges(keys=True)):
            try:
                value = link_attribute[link_name]

                if value < 0: # change the direction of the link and value
                    link_type = self[node1][node2][link_name]['type'] # 'type' should be the only other attribute on G.edge
                    self.remove_edge(node1, node2, link_name)
                    self.add_edge(node2, node1, link_name)
                    nx.set_edge_attributes(self, name='type', values={(node2, node1, link_name): link_type})
                    nx.set_edge_attributes(self, name='weight', values={(node2, node1, link_name): -value})
                else:
                    nx.set_edge_attributes(self, name='weight', values={(node1, node2, link_name): value})
            except:
                    pass

    def terminal_nodes(self):
        """
        Get all nodes with degree 1

        Parameters
        ----------
        G : graph
            A networkx graph

        Returns
        -------
        List of terminal node names
        
        """

        node_degree = dict(self.degree())
        terminal_nodes = [k for k,v in node_degree.items() if v == 1]

        return terminal_nodes

    def bridges(self):
        """
        Get bridge links (uses an undirected graph)

        Parameters
        ----------
        G : graph
            A networkx graph

        Returns
        -------
        List of links that are bridges
        
        """
        n = nx.number_connected_components(self.to_undirected())
        bridges = []
        for (node1, node2, link_name) in list(self.edges(keys=True)):
            # if node1 and node2 have a neighbor in common, no bridge
            if len(set(self.neighbors(node1)) & set(self.neighbors(node2))) == 0:
                self.remove_edge(node1, node2, key=link_name)
                if nx.number_connected_components(self.to_undirected()) > n:
                    bridges.append(link_name)
                self.add_edge(node1, node2, key=link_name)

        return bridges

    def central_point_dominance(self):
        """
        Compute central point dominance

        Returns
        -------
        Central point dominance (float)
        
        """
        bet_cen = nx.betweenness_centrality(self.to_undirected())
        bet_cen = list(bet_cen.values())
        cpd = sum(max(bet_cen) - np.array(bet_cen))/(len(bet_cen)-1)

        return cpd

    def spectral_gap(self):
        """
        Spectral gap. Difference in the first and second eigenvalue of
        the adjacency matrix

        Returns
        -------
        Spectral gap (float)
        
        """

        eig = nx.adjacency_spectrum(self.to_undirected())
        spectral_gap = abs(eig[0] - eig[1])

        return spectral_gap.real

    def algebraic_connectivity(self):
        """
        Algebraic connectivity. Second smallest eigenvalue of the normalized
        Laplacian matrix of a network. Uses an undirected graph.

        Returns
        -------
        Algebraic connectivity (float)
        
        """
        eig = nx.laplacian_spectrum(self.to_undirected())
        eig = np.sort(eig)
        alg_con = eig[1]

        return alg_con

    def critical_ratio_defrag(self):
        """
        Critical ratio of defragmentation.

        Returns
        -------
        Critical ratio of defragmentation (float)
        
        """
        node_degree = dict(self.degree())
        tmp = np.mean(pow(np.array(list(node_degree.values())),2))
        fc = 1-(1/((tmp/np.mean(list(node_degree.values())))-1))

        return fc

    def links_in_simple_paths(self, sources, sinks):
        """
        Count all links in a simple path between sources and sinks

        Parameters
        -----------
        sources : list
            List of source nodes
        sinks : list
            List of sink nodes
        sinks : list
            List of sink nodes

        Returns
        -------
        Dictionary with the number of times each link is involved in a path
        
        """
        link_names = [name for (node1, node2, name) in list(self.edges(keys=True))]
        link_count = pd.Series(data = 0, index=link_names)

        for sink in sinks:
            for source in sources:
                if nx.has_path(self, source, sink):
                    paths = _all_simple_paths(self,source,target=sink)
                    for path in paths:
                        for i in range(len(path)-1):
                            links = list(self[path[i]][path[i+1]].keys())
                            for link in links:
                                link_count[link] = link_count[link]+1

        return link_count


def _all_simple_paths(G, source, target, cutoff=None):
    """
    Adaptation of nx.all_simple_paths for multigraphs
    """

    if source not in G:
        raise nx.NetworkXError('source node %s not in graph'%source)
    if target not in G:
        raise nx.NetworkXError('target node %s not in graph'%target)
    if cutoff is None:
        cutoff = len(G)-1
    if G.is_multigraph():
        return _all_simple_paths_multigraph(G, source, target, cutoff=cutoff)
    else:
        return 1 #_all_simple_paths_graph(G, source, target, cutoff=cutoff)


def _all_simple_paths_multigraph(G, source, target, cutoff=None):
    if cutoff < 1:
        return
    visited = [source]
    stack = [(v for u,v,k in list(G.edges(source, keys=True)))]
    while stack:
        children = stack[-1]
        child = next(children, None)
        if child is None:
            stack.pop()
            visited.pop()
        elif nx.has_path(G, child, target) == False: # added kaklise
            pass
        elif len(visited) < cutoff:
            if child == target:
                yield visited + [target]
            elif child not in visited:
                visited.append(child)
                stack.append((v for u,v in list(G.edges(child))))
        else: #len(visited) == cutoff:
            count = ([child]+list(children)).count(target)
            for i in range(count):
                yield visited + [target]
            stack.pop()
            visited.pop()




"""
The wntr.morph package contains methods to modify water network 
model morphology, including network skeletonization, modifying 
node coordinates, and splitting or breaking pipes.
"""
from wntr.morph.node import scale_node_coordinates, translate_node_coordinates, \
    rotate_node_coordinates
from wntr.morph.link import split_pipe, break_pipe
from wntr.morph.skel import skeletonize


import logging
from wntr import aml

logger = logging.getLogger(__name__)


def demand_var(m, wn, index_over=None):
    """
    Add a demand variable to the model

    Parameters
    ----------
    m: wntr.aml.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    index_over: list of str
        list of junction names
    """
    if not hasattr(m, 'demand'):
        m.demand = aml.VarDict()

    if index_over is None:
        index_over = wn.junction_name_list

    for node_name in index_over:
        node = wn.get_node(node_name)
        m.demand[node_name] = aml.Var(node.demand_timeseries_list(wn.sim_time))


def flow_var(m, wn, index_over=None):
    """
    Add a flow variable to the model

    Parameters
    ----------
    m: wntr.aml.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    index_over: list of str
        list of link names
    """
    if not hasattr(m, 'flow'):
        m.flow = aml.VarDict()

    if index_over is None:
        index_over = wn.link_name_list

    for link_name in index_over:
        link = wn.get_link(link_name)
        m.flow[link_name] = aml.Var(0.001)


def head_var(m, wn, index_over=None):
    """
    Add a head variable to the model

    Parameters
    ----------
    m: wntr.aml.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    index_over: list of str
        list of junction names
    """
    if not hasattr(m, 'head'):
        m.head = aml.VarDict()

    if index_over is None:
        index_over = wn.junction_name_list

    for node_name in index_over:
        node = wn.get_node(node_name)
        m.head[node_name] = aml.Var(node.elevation)


def leak_rate_var(m, wn, index_over=None):
    """
    Add a variable to the model for leak flow rate

    Parameters
    ----------
    m: wntr.aml.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    index_over: list of str
        list of junction/tank names
    """
    if not hasattr(m, 'leak_rate'):
        m.leak_rate = aml.VarDict()

    if index_over is None:
        index_over = wn.junction_name_list + wn.tank_name_list

    for node_name in index_over:
        m.leak_rate[node_name] = aml.Var(0)





"""
The wntr.metrics package contains methods to compute resilience, including
hydraulic, water quality, water security, and economic metrics.  Methods to 
compute topographic metrics are included in the wntr.network.graph module.
"""
from wntr.metrics.hydraulic import expected_demand, average_expected_demand, water_service_availability, todini_index, entropy
from wntr.metrics.water_security import mass_contaminant_consumed, volume_contaminant_consumed, extent_contaminant
from wntr.metrics.economic import annual_network_cost, annual_ghg_emissions, pump_energy, pump_cost
from wntr.metrics.misc import query, population, population_impacted




"""
The wntr.epanet package provides EPANET2 compatibility functions for WNTR.
"""
from .io import InpFile  #, BinFile, HydFile, RptFile
from .util import FlowUnits, MassUnits, HydParam, QualParam, EN
import wntr.epanet.toolkit




"""
The wntr.morph.node module contains functions to modify node coordinates.
"""
import logging
import sys
import copy
import numpy as np
    
logger = logging.getLogger(__name__)


def _deepcopy_wn(wn):
    """
    Increase recursion limit for python 2.7 from 1000 to 3000
    """
    recursion_limit = sys.getrecursionlimit()
    
    if sys.version_info.major < 3:
        sys.setrecursionlimit(3000) 
        
    wn2 = copy.deepcopy(wn)
    
    if sys.version_info.major < 3:
        sys.setrecursionlimit(recursion_limit)
    
    return wn2

def scale_node_coordinates(wn, scale):
    """
    Scales node coordinates, using 1:scale
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        A WaterNetworkModel object
    
    scale: float
        Coordinate scale multiplier, in meters
    
    Returns
    --------
    A WaterNetworkModel object with updated node coordinates
    """
    wn2 = _deepcopy_wn(wn)
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        node.coordinates = (pos[0]*scale, pos[1]*scale)

    return wn2


def translate_node_coordinates(wn, offset_x, offset_y):
    """
    Translate node coordinates
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        A WaterNetworkModel object
    
    offset_x: tuple
        Translation in the x direction, in meters
    
    offset_y: float
        Translation in the y direction, in meters
    
    Returns
    --------
    A WaterNetworkModel object with updated node coordinates
    """
    wn2 = _deepcopy_wn(wn)
    
    for name, node in wn2.nodes():
        pos = node.coordinates
        node.coordinates = (pos[0]+offset_x, pos[1]+offset_y)
    
    return wn2
        

def rotate_node_coordinates(wn, theta):
    """
    Rotate node coordinates counter-clockwise by theta degrees
    
    Parameters
    -----------
    wn: wntr WaterNetworkModel
        A WaterNetworkModel object
    
    theta: float
        Node rotation, in degrees
    
    Returns
    --------
    A WaterNetworkModel object with updated node coordinates
    """
    wn2 = _deepcopy_wn(wn)
    
    theta = np.radians(theta)
    R = np.array([[np.cos(theta),-np.sin(theta)], 
                  [np.sin(theta), np.cos(theta)]])
    for name, node in wn2.nodes():
        pos = node.coordinates
        node.coordinates = tuple(np.dot(R,pos))
    
    return wn2



import numpy as np
import scipy.sparse as sp
import warnings
import logging
import enum

warnings.filterwarnings("error",'Matrix is exactly singular', sp.linalg.MatrixRankWarning)
np.set_printoptions(precision=3, threshold=10000, linewidth=300)

logger = logging.getLogger(__name__)


class SolverStatus(enum.IntEnum):
    converged = 1
    error = 0


class NewtonSolver(object):
    """
    Newton Solver class.
    """
    
    def __init__(self, options=None):
        if options is None:
            options = {}
        self._options = options

        if 'MAXITER' not in self._options:
            self.maxiter = 3000
        else:
            self.maxiter = self._options['MAXITER']

        if 'TOL' not in self._options:
            self.tol = 1e-6
        else:
            self.tol = self._options['TOL']

        if 'BT_RHO' not in self._options:
            self.rho = 0.5
        else:
            self.rho = self._options['BT_RHO']

        if 'BT_MAXITER' not in self._options:
            self.bt_maxiter = 100
        else:
            self.bt_maxiter = self._options['BT_MAXITER']

        if 'BACKTRACKING' not in self._options:
            self.bt = True
        else:
            self.bt = self._options['BACKTRACKING']

        if 'BT_START_ITER' not in self._options:
            self.bt_start_iter = 0
        else:
            self.bt_start_iter = self._options['BT_START_ITER']

        if 'THREADS' not in self._options:
            self.num_threads = 4
        else:
            self.num_threads = self._options['THREADS']

    def solve(self, model):
        """

        Parameters
        ----------
        model: wntr.aml.Model

        Returns
        -------
        status: SolverStatus
        message: str
        """
        logger_level = logger.getEffectiveLevel()

        x = model.get_x()
        if len(x) == 0:
            return SolverStatus.converged, 'No variables or constraints'

        use_r_ = False

        # MAIN NEWTON LOOP
        for outer_iter in range(self.maxiter):
            if use_r_:
                r = r_
                r_norm = new_norm
            else:
                r = model.evaluate_residuals(num_threads=self.num_threads)
                r_norm = np.max(abs(r))

            if logger_level <= 1:
                if outer_iter < self.bt_start_iter:
                    logger.log(1, 'iter: {0:<4d} norm: {1:<10.2e}'.format(outer_iter, r_norm))

            if r_norm < self.tol:
                return SolverStatus.converged, 'Solved Successfully'

            J = model.evaluate_jacobian(x=None)

            # Call Linear solver
            try:
                d = -sp.linalg.spsolve(J, r, permc_spec='COLAMD', use_umfpack=False)
            except sp.linalg.MatrixRankWarning:
                return SolverStatus.error, 'Jacobian is singular at iteration ' + str(outer_iter)

            # Backtracking
            alpha = 1.0
            if self.bt and outer_iter >= self.bt_start_iter:
                use_r_ = True
                for iter_bt in range(self.bt_maxiter):
                    x_ = x + alpha*d
                    model.load_var_values_from_x(x_)
                    r_ = model.evaluate_residuals(num_threads=self.num_threads)
                    new_norm = np.max(abs(r_))
                    if new_norm < (1.0-0.0001*alpha)*r_norm:
                        x = x_
                        break
                    else:
                        alpha = alpha*self.rho

                if iter_bt+1 >= self.bt_maxiter:
                    return SolverStatus.error, 'Line search failed at iteration ' + str(outer_iter)
                if logger_level <= 1:
                    logger.log(1, 'iter: {0:<4d} norm: {1:<10.2e} alpha: {2:<10.2e}'.format(outer_iter, new_norm, alpha))
            else:
                x += d
                model.load_var_values_from_x(x)
            
        return SolverStatus.error, 'Reached maximum number of iterations: ' + str(outer_iter)






from wntr.utils.ordered_set import OrderedDict, OrderedSet
from six import with_metaclass
import abc


class ModelUpdater(object):
    def __init__(self):
        self.update_functions = OrderedDict()

    def add(self, obj, attr, func):
        if (obj, attr) not in self.update_functions:
            self.update_functions[(obj, attr)] = OrderedSet()
        self.update_functions[(obj, attr)].add(func)

    def update(self, m, wn, obj, attr):
        if (obj, attr) in self.update_functions:
            for func in self.update_functions[(obj, attr)]:
                func(m, wn, self, obj, attr)


class Definition(with_metaclass(abc.ABCMeta, object)):
    @classmethod
    @abc.abstractmethod
    def build(cls, m, wn, updater, index_over=None):
        pass

    @classmethod
    def update(cls, m, wn, updater, obj, attr):
        cls.build(m, wn, updater, index_over=[obj.name])





import sys
if sys.version_info.major == 2:
    from collections import MutableSet
else:
    from collections.abc import MutableSet
from collections import OrderedDict
from collections import Iterable


class OrderedSet(MutableSet):
    """
    An ordered set.
    """
    def __init__(self, iterable=None):
        """
        Parameters
        ----------
        iterable: Iterable
            An iterable with wich to initialize the set.
        """
        self._data = OrderedDict()
        if iterable is not None:
            self.update(iterable)

    def __contains__(self, item):
        return item in self._data

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return len(self._data)

    def add(self, value):
        """
        Add an element to the set.

        Parameters
        ----------
        value: object
            The object to be added to the set.
        """
        self._data[value] = None

    def discard(self, value):
        """
        Discard and element from the set.

        Parameters
        ----------
        value: object
            The object to be discarded.
        """
        self._data.pop(value, None)

    def update(self, iterable):
        """
        Update the set with the objects in iterable.

        Parameters
        ----------
        iterable: Iterable
        """
        for i in iterable:
            self.add(i)

    def __repr__(self):
        s = '{'
        for i in self:
            s += str(i)
            s += ', '
        s += '}'
        return s

    def __str__(self):
        return self.__repr__()

    def union(self, iterable):
        ret = OrderedSet(self)
        for i in iterable:
            ret.add(i)
        return ret

    def __sub__(self, other):
        ret = OrderedSet(self)
        for i in other:
            ret.discard(i)
        return ret
